﻿
using System;
using System.Drawing;
using System.Windows.Forms;

/* Interpolation by periodic cubic splines in x and y direction                  */
/* us of a Cholesky decomposation to solve the matrix equation                   */
/* www.mosismath.com                                                             */

namespace WindowsFormsApplication1
{

    public partial class MainWin : Form
    {
        public const int order = 10; // number of reference points (max 10)
        public const int datapoints = 1000; // number of datapoints for the interpolation
        public const double deltaX = 0.1; // sample time 

        public class TMatrix
        {
            public double[,] a;
            public double[] y;
            public double[] x;
            public double[] t;

            public TMatrix(int size)
            {
                a = new double[size, size];
                y = new double[size];
                x = new double[size];
                t = new double[size];
            }
        }

        TMatrix m;
        double[] yp = new double[datapoints];
        double[] xp = new double[datapoints];
        double[] x_k = new double[order+1];
        double[] y_k = new double[order + 1];
        double[] t_k = new double[order+1];
        double[] ax = new double[order+1];
        double[] bx = new double[order+1];
        double[] cx = new double[order+1];
        double[] dx = new double[order+1];
        double[] h = new double[order];

        double[] ay = new double[order + 1];
        double[] by = new double[order + 1];
        double[] cy = new double[order + 1];
        double[] dy = new double[order + 1];

        public class TCholesky
        {

            public int maxOrder = order;
            TMatrix m;
            double[,] l;
            double[] y;

            public TCholesky(int size, TMatrix mi)
            {
                maxOrder = size;
                l = new double[maxOrder, maxOrder]; // lower diagonal Matrix
                y = new double[maxOrder];
                m = mi;
            }

            public bool Decomp()
            {
                int i, j, k;
                bool result = true;

                for (i = 0; i < maxOrder; i++)
                {
                    for (j = 0; j <= i; j++)
                    {
                        if (i > j)
                        {
                            l[i, j] = m.a[i, j];
                            for (k = 0; k < j; k++)
                                l[i, j] = l[i, j] - l[i, k] * l[j, k];

                            if (Math.Abs(l[j, j]) > 1E-200)
                                l[i, j] = l[i, j] / l[j, j];
                            else
                                result = false;
                        }
                        if (i == j)
                        {
                            l[i, i] = m.a[i, i];
                            for (k = 0; k < j; k++)
                            {
                                l[i, i] = l[i, i] - (l[i, k] * l[i, k]);
                            }
                            if (l[i, i] >= 0)
                                l[i, i] = Math.Sqrt(l[i, i]);
                            else
                                result = false;
                        }
                    }
                }
                return result;
            }

            public bool Solve()
            {
                bool result = true;
                int i, j;
                for (j = 0; j < maxOrder; j++)
                {
                    y[j] = m.y[j];
                    for (i = 0; i < j; i++)
                    {
                        y[j] = y[j] - y[i] * l[j, i];
                    }
                    if (Math.Abs(l[j, j]) > 1E-200)
                        y[j] = y[j] / l[j, j];
                    else
                        result = false;
                }
                m.x[maxOrder - 1] = y[maxOrder - 1] / l[maxOrder - 1, maxOrder - 1];

                for (j = maxOrder - 2; j >= 0; j--)
                {
                    m.x[j] = y[j];
                    for (i = maxOrder - 1; i > j; i--)
                    {
                        m.x[j] = m.x[j] - m.x[i] * l[i, j];
                    }
                    if (Math.Abs(l[j, j]) > 1E-200)
                        m.x[j] = m.x[j] / l[j, j];
                    else
                        result = false;
                }
                return result;
            }

        }

        TCholesky cholesky;

        public MainWin()
        {
            int i;
            m = new TMatrix(order-1);
            cholesky = new TCholesky(order - 1, m);
            DataGridViewCell cell;

            InitializeComponent();

            dGPolynom.ColumnCount = order;
            dGPolynom.RowCount = 2;

            // init some sample values
            y_k[0] = 6.0;
            y_k[1] = 9.5;
            y_k[2] = 7.0;
            y_k[3] = 4.0;
            y_k[4] = 3.0;
            y_k[5] = 5.0;
            y_k[6] = 6.0;
            y_k[7] = 4.0;
            y_k[8] = 2.0;
            y_k[9] = 6.0;

            x_k[0] = 9.0;
            x_k[1] = 6.0;
            x_k[2] = 3.0;
            x_k[3] = 2.0;
            x_k[4] = 4.0;
            x_k[5] = 5.0;
            x_k[6] = 3.0;
            x_k[7] = 1.0;
            x_k[8] = 4.0;
            x_k[9] = 9.0;

            // 1timestamp to t value
            for (i = 1; i <= order; i++)
            {
                t_k[i - 1] = deltaX * (double)i;
            }

            cell = dGPolynom.Rows[0].HeaderCell;
            cell.Value = "y_k";
            cell = dGPolynom.Rows[1].HeaderCell;
            cell.Value = "x_k";
            for (i = 0; i < order; i++)
            {
                dGPolynom.Columns[i].Name = "a" + i.ToString();
                cell = dGPolynom[i, 0];
                cell.Value = Convert.ToString(y_k[i]);
                cell = dGPolynom[i, 1];
                cell.Value = Convert.ToString(x_k[i]);
            }
        }

        public bool CalcParameters()
        {
            int i, k;
            for (i = 0; i < order; i++)  // calculate a parameters
            {
                ax[i] = x_k[i];
                ay[i] = y_k[i];
            }
            ax[order] = x_k[1];
            ay[order] = y_k[1];
            for (i = 0; i < order-1; i++)
            {
                h[i] = t_k[i + 1] - t_k[i];
            }
            h[i] = h[0];
            

            // empty matrix for c parameters
            for (i = 0; i < order - 1; i++)
            {
                for (k = 0; k < order - 1; k++)
                {
                    m.a[i, k] = 0.0;
                    m.y[i] = 0.0;
                    m.x[i] = 0.0;
                }
            }

            // fill matrix for cx parameters
            for (i = 0; i < order - 1; i++)
            {
                if (i == 0)
                {
                    m.a[i, 0] = 2.0 * (h[0] + h[1]);
                    m.a[i, 1] = h[1];
                }
                else
                {
                    m.a[i, i - 1] = h[i];
                    m.a[i, i] = 2.0 * (h[i] + h[i + 1]);
                    if (i < order - 2)
                        m.a[i, i + 1] = h[i + 1];
                }
                if ((h[i] != 0.0) && (h[i + 1] != 0.0))
                    m.y[i] = ((ax[i + 2] - ax[i + 1]) / h[i + 1] - (ax[i + 1] - ax[i]) / h[i]) * 3.0;
                else
                    m.y[i] = 0.0;
            }
            m.a[0, order - 2] = h[0];
            m.a[order - 2, 0] = h[0];

            if (cholesky.Decomp())
            {
                cholesky.Solve();

                for (i = 1; i < order; i++)
                {
                    cx[i] = m.x[i - 1];
                }
                cx[0] = cx[order - 1];  // boundry condition for natural spline 
                for (i = 0; i < order; i++) // calculate b and d parameters
                {
                    if (h[i] != 0.0)
                    {
                        dx[i] = 1.0 / 3.0 / h[i] * (cx[i + 1] - cx[i]);
                        bx[i] = 1.0 / h[i] * (ax[i + 1] - ax[i]) - h[i] / 3.0 * (cx[i + 1] + 2 * cx[i]);
                    }
                }

                // empty matrix for cy parameters
                for (i = 0; i < order - 1; i++)
                {
                    for (k = 0; k < order - 1; k++)
                    {
                        m.a[i, k] = 0.0;
                        m.y[i] = 0.0;
                        m.x[i] = 0.0;
                    }
                }
                // fill matrix for c parameters
                for (i = 0; i < order - 1; i++)
                {
                    if (i == 0)
                    {
                        m.a[i, 0] = 2.0 * (h[0] + h[1]);
                        m.a[i, 1] = h[1];
                    }
                    else
                    {
                        m.a[i, i - 1] = h[i];
                        m.a[i, i] = 2.0 * (h[i] + h[i + 1]);
                        if (i < order - 2)
                            m.a[i, i + 1] = h[i + 1];
                    }
                    if ((h[i] != 0.0) && (h[i + 1] != 0.0))
                        m.y[i] = ((ay[i + 2] - ay[i + 1]) / h[i + 1] - (ay[i + 1] - ay[i]) / h[i]) * 3.0;
                    else
                        m.y[i] = 0.0;
                }
                m.a[0, order - 2] = h[0];
                m.a[order - 2, 0] = h[0];

                if (cholesky.Decomp())
                {
                    cholesky.Solve();

                    for (i = 1; i < order; i++)
                    {
                        cy[i] = m.x[i - 1];
                    }
                    cy[0] = cy[order - 1];  // boundry condition for natural spline 
                    for (i = 0; i < order; i++) // calculate b and d parameters
                    {
                        if (h[i] != 0.0)
                        {
                            dy[i] = 1.0 / 3.0 / h[i] * (cy[i + 1] - cy[i]);
                            by[i] = 1.0 / h[i] * (ay[i + 1] - ay[i]) - h[i] / 3.0 * (cy[i + 1] + 2 * cy[i]);
                        }
                    }
                    return true;
                }
                else
                    return false;
            }
            else
                return false;
        }

        public int Interpolate(int order, int resolution)
        {
            int i, k;
            double timestamp;
            for (i = 0; i < order - 1; i++)
            {
                for (k = 0; k < resolution; k++)
                {
                    timestamp = (double)(k) / (double)(resolution) * h[i];
                    xp[i * resolution + k] = ax[i] + bx[i] * (timestamp) + cx[i] * Math.Pow(timestamp, 2) + dx[i] * Math.Pow(timestamp, 3);
                    yp[i * resolution + k] = ay[i] + by[i] * (timestamp) + cy[i] * Math.Pow(timestamp, 2) + dy[i] * Math.Pow(timestamp, 3);
                }
            }
            xp[i * resolution] = xp[0];
            yp[i * resolution] = yp[0];
            return i * resolution;
        }

        public void DrawGraph(int maxPoints, bool doClear)
        {
            Point p1, p2;
            int j;
            double maxTime = -1.0;
            double maxVal  = -1.0;
            double minVal  = 1.0;
            double scalefactor;
            bool drawCeroline;
            for (j = 0; j < maxPoints - 1; j++)
            {
                if (maxTime < xp[j])
                    maxTime = xp[j];
                if (maxVal < yp[j])
                    maxVal = yp[j];
                if (minVal > yp[j])
                    minVal = yp[j];
            }
            maxTime = maxTime * 1.05;
            if (minVal < 0)
            {
                if (maxVal > 0)
                {
                    if (maxVal > Math.Abs(minVal))
                        scalefactor = pGraph.Height / maxVal / 2.2;
                    else
                        scalefactor = pGraph.Height / Math.Abs(minVal) / 2.2;
                }
                else
                    scalefactor = 5.0;
                drawCeroline = true;
            }
            else
            {
                scalefactor = pGraph.Height / (maxVal) / 1.1;
                drawCeroline = false;
            }
            p1 = new Point();
            p2 = new Point();
            Graphics g = pGraph.CreateGraphics();
            if (doClear)
                g.Clear(Color.White);
            Pen bluePen = new Pen(Color.Blue, 2);
            Pen redPen = new Pen(Color.Red, 2);
            Pen blackPen = new Pen(Color.Black, 2);
            if (drawCeroline)
            {
                p1.X = 0;
                p1.Y = (pGraph.Height / 2);
                p2.X = pGraph.Width;
                p2.Y = (pGraph.Height / 2);
                g.DrawLine(blackPen, p1, p2);
            }
            bluePen.Width = 1;
            if (maxTime > 0) // draw interpolated graph
            {
                for (j = 0; j < maxPoints - 1; j++)
                {
                    p1.X =  Convert.ToInt32(Math.Round(xp[j] * pGraph.Width / (double)maxTime));
                    p2.X =  Convert.ToInt32(Math.Round(xp[j + 1] * pGraph.Width / (double)maxTime));
                    if (drawCeroline)
                    {
                        p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(yp[j] * scalefactor));
                        p2.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(yp[j + 1] * scalefactor));
                    }
                    else
                    {
                        p1.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(yp[j] * scalefactor));
                        p2.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(yp[j + 1] * scalefactor));
                    }
                   
                        g.DrawLine(bluePen, p1, p2);
                }
                for (j = 0; j < order; j++) // mark reference points
                {
                    p1.X = Convert.ToInt32(x_k[j] * pGraph.Width / (double)maxTime);
                    if (drawCeroline)
                        p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(y_k[j] * scalefactor));
                    else
                        p1.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(y_k[j] * scalefactor));
                    g.DrawEllipse(redPen, (int)(p1.X - 1.0), (int)(p1.Y - 1.0), 2, 2);
                }
            }
            g.Dispose();  
        }

        private void button1_MouseClick(object sender, MouseEventArgs e)
        {
            int i;
            DataGridViewCell cell;

            for (i = 0; i < order; i++)
            {
                cell = dGPolynom[i, 0];
                if (cell.ToString() != "")
                {
                    y_k[i] = Convert.ToDouble(cell.Value);
                    cell = dGPolynom[i, 1];
                    x_k[i] = Convert.ToDouble(cell.Value);
                }
                else
                {
                    y_k[i] = 0.0;
                    x_k[i] = 0.0;
                }
            }
            if (CalcParameters())
            {
                DrawGraph(Interpolate(order, datapoints / (order - 1)), true);
            }
            else
                MessageBox.Show("Matrix calculattion", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
        }


        private void pGraph_Paint(object sender, PaintEventArgs e)
        {
            DrawGraph(datapoints, false);
        }

    }
}
